# -*- coding: utf-8 -*-
import os

NSBL_MODULE_FOLDER = os.path.dirname(__file__)

TEMPLATES_FOLDER = os.path.join(NSBL_MODULE_FOLDER, "templates")

# stem key for inventory
ENVS_KEY = "envs"
# meta info for groups/hosts (contains for example 'hosts' for groups)
ENV_META_KEY = "meta"
# name of the group/host
ENV_NAME_KEY = "name"
# type of the environemnt (either group or host)
ENV_TYPE_KEY = "type"
# under meta key, lists hosts of a group
ENV_HOSTS_KEY = "hosts"
# under meta key, lists sub-groups of a group, or which groups a host is member is
ENV_GROUPS_KEY = "groups"
# vars for a host/group
VARS_KEY = "vars"
# tasks for a hosts/group, used to create playbooks
TASKS_KEY = "tasks"

ENV_TYPE_HOST = "host"
ENV_TYPE_GROUP = "group"
DEFAULT_ENV_TYPE = ENV_TYPE_GROUP

# id of the environment tasks are run in
ENV_ID_KEY = "_env_id"

TASKS_META_KEY = "meta"

# whether to 'import_*' or 'include_*' roles & tasks if not specified
DEFAULT_INCLUDE_TYPE = "include"

ADD_TYPE_TASK_LIST = "TASK_LIST"
ADD_TYPE_TASK_LIST_FILE = "TASK_LIST_FILE"
ADD_TYPE_ROLE = "ROLE"
ADD_TYPE_CALLBACK = "CALLBACK_PLUGIN"
ADD_TYPE_ACTION = "ACTION_PLUGIN"
ADD_TYPE_LIBRARY = "LIBRARY_PLUGIN"
ADD_TYPE_FILTER = "FILTER_PLUGIN"
ADD_TYPE_MODULE_UTIL = "MODULE_UTIL"
ADD_TYPE_FILES = "FILES"
ADD_FILE_TYPES = [
    ADD_TYPE_TASK_LIST,
    ADD_TYPE_ROLE,
    ADD_TYPE_CALLBACK,
    ADD_TYPE_ACTION,
    ADD_TYPE_LIBRARY,
    ADD_TYPE_FILTER,
    ADD_TYPE_MODULE_UTIL,
    ADD_TYPE_FILES,
    ADD_TYPE_TASK_LIST_FILE,
]

ANSIBLE_ROLE_CACHE_DIR = os.path.expanduser("~/.cache/ansible-roles")

# only used for output
ROLE_ID_KEY = "_role_id"
TASK_META_NAME_KEY = "name"
DYN_TASK_ID_KEY = "_dyn_task_id"
DYN_ROLE_TYPE = "dyn_role"

MITOGEN_VERSION = "0.2.9"


# from freckles, make sure to keep in sync
FRECKLES_DESC_METADATA_KEY = "desc"
FRECKLES_DESC_SHORT_METADATA_KEY = "short"
FRECKLES_DESC_LONG_METADATA_KEY = "long"
FRECKLES_DESC_REFERENCES_METADATA_KEY = "references"
FRECKLES_PROPERTIES_METADATA_KEY = "properties"
FRECKLES_PROPERTIES_IDEMPOTENT_METADATA_KEY = "idempotent"
FRECKLES_PROPERTIES_INTERNET_METADATA_KEY = "internet"
FRECKLES_PROPERTIES_ELEVATED_METADATA_KEY = "elevated"



# -*- coding: utf-8 -*-

import logging
import sys

import click
from six import string_types

log = logging.getLogger("nsbl")

ENCODING = sys.stdout.encoding
if not ENCODING:
    ENCODING = "utf-8"

#
# def get_terminal_width():
#     command = ["tput", "cols"]
#     try:
#         width = int(subprocess.check_output(command))
#     except OSError:
#         return -1
#     except subprocess.CalledProcessError:
#         return -1
#     else:
#         return width


class NsblPrintCallbackAdapter(object):
    def __init__(self):

        self.environment_parameters = None

    def set_environment_parameters(self, parameters):

        self.environment_parameters = parameters

    def add_error_message(self, line):

        click.echo(line.strip())

    def add_log_message(self, line):

        click.echo(line)

    def write(self, line):

        if line.strip():
            # self.current_lines.append(line)
            self.add_log_message(line)

    def flush(self):

        pass

    def finish_up(self):
        pass


def translate_item_to_task_name(action, item):

    name = None
    if action == "install":
        if isinstance(item, string_types):
            return item

        name = item.get("name", None)
        if name is None:
            name = item.get("requirements", None)
        if name is None:
            name = item.get("repo", None)
        if name is None:
            name = item.get("url", None)
        if name is None:
            name = item.get("src", None)

    if name is None:
        name = item

    return name



# -*- coding: utf-8 -*-

from __future__ import absolute_import, division, print_function, unicode_literals

import copy
import io
import logging
import os
import signal
import subprocess
import sys
import time
from secrets import token_urlsafe

import cursor
from future import standard_library

try:
    from shutil import which
except Exception:
    # python2
    from distutils.spawn import find_executable as which
from plumbum import local

from frutils import readable_yaml
from .nsbl_callback import NsblPrintCallbackAdapter

standard_library.install_aliases()


log = logging.getLogger("nsbl")

DEFAULT_PEXPECT_TIMEOUT = 36000


class NsblRunner(object):
    def __init__(self, nsbl):
        """Class to kick off rendering and running the ansible environment in question.

        Args:
          nsbl (Nsbl): the Nsbl object holding the (processed) configuration
        """
        self.nsbl = nsbl

    def process_secret_vars_env(
        self,
        run_folder,
        run_env,
        sudo_password,
        ssh_password,
        secret_vars,
        elevated_permissions_required,
        passwordless_sudo_possible,
    ):

        # adding secure_vars
        for sec_var_alias, sec_var_details in secret_vars.items():
            if sec_var_details["type"] == "environment":
                run_env[sec_var_alias] = sec_var_details["value"]
            else:
                raise Exception(
                    "Secure var type '{}' not supported.".format(
                        sec_var_details["type"]
                    )
                )

        if sudo_password is not None:
            if elevated_permissions_required:
                if passwordless_sudo_possible:
                    ask_sudo = False
                else:
                    ask_sudo = True
            else:
                ask_sudo = False
        else:
            ask_sudo = False

        if ask_sudo:
            run_env["NSBL_SUDO_PASSWORD"] = sudo_password
        if ssh_password is not None:
            run_env["NSBL_SSH_PASSWORD"] = ssh_password

    def process_secret_vars_vault(
        self,
        run_folder,
        run_env,
        encryption_key,
        sudo_password,
        ssh_password,
        secret_vars,
        elevated_permissions_required,
        passwordless_sudo_possible,
    ):

        vars_dict = copy.copy(secret_vars)
        if sudo_password is not None:
            if elevated_permissions_required:
                if passwordless_sudo_possible:
                    ask_sudo = False
                else:
                    ask_sudo = True
            else:
                ask_sudo = False
        else:
            ask_sudo = False

        if ask_sudo:
            vars_dict["ansible_become_pass"] = sudo_password
        if ssh_password is not None:
            vars_dict["ansible_ssh_pass"] = ssh_password

        # create_vault_script_path = os.path.join(
        #     os.path.dirname(__file__), "external", "scripts", "create-vault.sh"
        # )
        create_vault_script_path = os.path.join(run_folder, "create-vault.sh")
        create_vault_script = local[create_vault_script_path]

        play_folder = os.path.join(run_folder, "plays")

        process = None
        vault_pipe = os.path.join(play_folder, "unencrypted_vault")
        pw_file_pipe = os.path.join(play_folder, "vault_pw")
        try:
            # os.mkfifo(pw_file_pipe, mode=0o700)
            # os.mkfifo(vault_pipe, mode=0o700)
            with io.open(pw_file_pipe, "w+") as fh:
                fh.write("")
            os.chmod(pw_file_pipe, 0o600)
            with io.open(vault_pipe, "w+") as fh:
                fh.write("")
            os.chmod(vault_pipe, 0o600)

            with io.open(pw_file_pipe, "w", encoding="UTF-8") as f:
                f.write("{}\n".format(encryption_key))

            with io.open(vault_pipe, "w", encoding="UTF-8") as f:
                f.write("{}\n".format(readable_yaml(vars_dict, ignore_aliases=True)))

            # rc, stdout, stderr = create_vault_script(play_folder)
            process = create_vault_script.popen([play_folder], env=run_env)
            process.wait()

            rc = process.returncode

            if rc != 0:
                raise Exception("Return code of 'create-vault.sh' script not 0")
        except (Exception) as e:
            vault_file = os.path.join(play_folder, "run_vault.yml")
            if os.path.exists(vault_file):
                os.unlink(vault_file)

            raise Exception("Could not create vault: {}".format(e))
        finally:
            if os.path.exists(pw_file_pipe):
                os.unlink(pw_file_pipe)
            if os.path.exists(vault_pipe):
                os.unlink(vault_pipe)

    def run(
        self,
        run_folder=None,
        global_vars=None,
        force=False,
        ansible_args="",
        elevated_permissions_required=None,
        passwordless_sudo_possible=None,
        sudo_password=None,
        ssh_password=None,
        extra_env_vars=None,
        secure_vars=None,
        callback=None,
        callback_adapter=None,
        add_timestamp_to_env=False,
        add_symlink_to_env=False,
        no_run=False,
        pre_run_callback=None,
        extra_paths=None,
        show_tasks_with_password_in_log=False,
        use_mitogen=False,
        use_ara=False,
        parent_task=None,
        **kwargs
    ):
        """Starts the ansible run, executing all generated playbooks.

        By default the 'nsbl_internal' ansible callback is used, which outputs easier to read outputs/results. You can, however,
        also use the callbacks that come with ansible, like 'default', 'skippy', etc.

        Args:
          run_folder (str): the target directory where the ansible environment should be rendered
          global_vars (dict): vars to be rendered on top of each playbook
          force (bool): whether to overwrite potentially existing files at the target (most likely an old rendered ansible environment)
          ansible_args (str): verbosity arguments to ansible-playbook command
          elevated_permissions_required (bool): whether to include the '--ask-become-pass' arg to the ansible-playbook call
          passwordless_sudo_possible (bool): whether passwordless sudo is possible on the host(s)
          sudo_password (str): if provided, it will be used instead of asking for a password
          ssh_password (str): if provided, it will be used instead of asking for a password
          extra_env_vars (dict): a dict with vars to put into the run environment
          secure_vars (dict): other vars to keep secure
          callback (str): the callback to use for the ansible run. default is 'default'
          callback_adapter: the callback adapter to use
          add_timestamp_to_env (bool): whether to append a timestamp to the run directory (default: False)
          add_symlink_to_env (str): whether to add a symlink to the run directory (will be deleted if exists already and force is specified) - default: False, otherwise path to symlink
          no_run (bool): whether to only render the environment, but not run it
          pre_run_callback (function): a callback to execute after the environment is rendered, but before the run is kicked off
          extra_paths (str): a list of extra paths that should be exported for the nsbl run
          show_tasks_with_password_in_log (bool): whether to show tasks that contain a password variable in the log (only enable for debugging!)
          use_mitogen (bool): whether to use mitogen to speed up playbook run
          use_ara (bool): use ara (https://ara.readthedocs.io)
          parent_task (TaskDetail): task object to update
        Return:
          dict: the parameters of the run
        """

        env_dir = None

        if run_folder is None:
            run_folder = os.path.join(os.getcwd(), "nsbl_env")

        if callback is None:
            callback = "default"

        if callback_adapter is None:
            callback_adapter = NsblPrintCallbackAdapter()

        # debug_callback = False
        # if debug_callback:
        #     callback_adapter = NsblPrintCallbackAdapter()
        #     callback = "freckles_callback"
        # else:
        #     if isinstance(callback, string_types):
        #         callback_adapter = NsblPrintCallbackAdapter()
        #     else:
        #         callback_adapter = FrecklesCallbackAdapter(self.nsbl, callback)
        #         callback = "freckles_callback"

        # try:
        #     term = Terminal()
        # except (Exception):
        #     # no terminal then, fine...
        #     # probably because the TERM env var is not set
        #     term = None

        try:
            parameters = None
            proc = None
            parameters = self.nsbl.render(
                run_folder,
                global_vars=global_vars,
                extract_vars=True,
                force=force,
                elevated_permissions_required=elevated_permissions_required,
                passwordless_sudo_possible=passwordless_sudo_possible,
                secure_vars=secure_vars,
                ansible_args=ansible_args,
                callback=callback,
                add_timestamp_to_env=add_timestamp_to_env,
                add_symlink_to_env=add_symlink_to_env,
                extra_paths=":".join(extra_paths),
                show_tasks_with_password_in_log=show_tasks_with_password_in_log,
                use_mitogen=use_mitogen,
                use_ara=use_ara,
                parent_task=parent_task,
            )
            callback_adapter.set_environment_parameters(parameters)

            env_dir = parameters["env_dir"]
            if pre_run_callback:
                pre_run_callback(env_dir)

            if no_run:
                log.debug("Not running environment due to 'no_run' flag set.")
                return parameters

            run_env = os.environ.copy()
            if callback.startswith("nsbl_internal"):
                run_env["NSBL_ENVIRONMENT"] = "true"

            # try to set 'USER' environment var, as that is sometimes not set in a docker environment
            if "USER" not in run_env.keys():
                whoami = which("whoami")

                if whoami:
                    which_cmd = local[whoami]
                    rc, stdout, stderr = which_cmd.run()

                    if rc == 0:
                        run_env["USER"] = stdout

            def preexec_function():
                # Ignore the SIGINT signal by setting the handler to the standard
                # signal handler SIG_IGN.
                signal.signal(signal.SIGINT, signal.SIG_IGN)

            script = parameters["run_playbooks_script"]

            path = os.getenv("PATH")
            path = "{}:{}".format(":".join(extra_paths), path)
            log.debug("PATHS for Ansible run: {}".format(path))

            if extra_env_vars is not None:
                for k, v in extra_env_vars.items():
                    run_env[k] = v

            # check if frozen
            env = dict(os.environ)  # make a copy of the environment
            lp_key = "LD_LIBRARY_PATH"  # for Linux and *BSD.
            lp_orig = env.get(lp_key + "_ORIG")  # pyinstaller >= 20160820 has this
            if lp_orig is not None:
                run_env[lp_key] = lp_orig  # restore the original, unmodified value
            else:
                run_env.pop(lp_key, None)  # last resort: remove the env var

            enc_key = token_urlsafe(32)

            self.process_secret_vars_vault(
                run_folder=run_folder,
                encryption_key=enc_key,
                sudo_password=sudo_password,
                ssh_password=ssh_password,
                secret_vars=secure_vars,
                elevated_permissions_required=elevated_permissions_required,
                run_env=run_env,
                passwordless_sudo_possible=passwordless_sudo_possible,
            )

            # write password file
            key_file = os.path.join(env_dir, "plays", "vault_pw")
            # local["touch"]([key_file])
            # print("TOUCHED: {}".format(key_file))
            with io.open(key_file, "w+") as fh:
                fh.write("")
            os.chmod(key_file, 0o600)
            with io.open(key_file, "w+") as fh:
                fh.write("{}".format(enc_key))

            # if sudo_password is None:
            proc = subprocess.Popen(
                script,
                stdout=subprocess.PIPE,
                stderr=sys.stdout.fileno(),
                stdin=subprocess.PIPE,
                shell=True,
                env=run_env,
                # preexec_fn=preexec_function,
                preexec_fn=os.setsid,
            )

            # proc.stdin.write(enc_key.encode(encoding="UTF-8"))
            # proc.stdin.close()

            for line in iter(proc.stdout.readline, ""):
                if line:
                    line = line.decode("utf-8").strip()
                    callback_adapter.add_log_message(line)
                if not line and proc.poll() is not None:
                    break

            while proc.poll() is None:
                # Process hasn't exited yet, let's wait some
                time.sleep(0.5)

            # Get return code from process
            return_code = proc.returncode
            parameters["return_code"] = return_code
            parameters["signal_status"] = -1

        except KeyboardInterrupt:
            cursor.show()  # looks like sometimes the cursor doesn't show otherwise...
            parameters["return_code"] = 11
            parameters["signal_status"] = -1

            # print()
            # print()
            # callback_adapter.add_error_message(
            #     "Keyboard interrupt received. Exiting..."
            # )
            # print()
            if proc is not None:
                os.killpg(os.getpgid(proc.pid), signal.SIGTERM)

            callback_adapter.cancel(
                "Keyboard interrupt received. Exiting. Already started tasks might or might not finish in the background..."
            )

        except (Exception) as e:

            if parameters is not None:
                parameters["return_code"] = 11
                parameters["signal_status"] = -1

            # import traceback

            # traceback.print_exc()
            # callback_adapter.add_error_message("Exception in Ansible run: {}".format(e))
            log.debug(e, exc_info=1)
            if proc is not None:
                os.killpg(os.getpgid(proc.pid), signal.SIGTERM)

            callback_adapter.cancel("Execution error: {}. Exiting...".format(e))
            raise e
        finally:
            callback_adapter.finish_up()

            if env_dir:
                run_vault_file = os.path.join(env_dir, "plays", "run_vault.yml")
                if os.path.exists(run_vault_file):
                    os.unlink(run_vault_file)
                vault_pw = os.path.join(env_dir, "plays", "vault_pw")
                if os.path.exists(vault_pw):
                    os.unlink(vault_pw)

        return {"nsbl_properties": parameters}



# -*- coding: utf-8 -*-

# python 3 compatibility
from __future__ import absolute_import, division, print_function, unicode_literals

import logging
import os

from ruamel.yaml.comments import CommentedSeq
from six import string_types

from frkl import load_object_from_url_or_path
from frutils import StringYAML, DEFAULT_EXCLUDE_DIRS
from .exceptions import NsblException

log = logging.getLogger("nsbl")

yaml = StringYAML()
yaml.default_flow_style = False

ANSIBLE_TASK_KEYWORDS = [
    "any_errors_fatal",
    "async",
    "become",
    "become_flags",
    "become_method",
    "become_user",
    "changed_when",
    "check_mode",
    "connection",
    "debugger",
    "delay",
    "delegate_facts",
    "delegate_to",
    "diff",
    "environment",
    "failed_when",
    "ignore_errors",
    "loop",
    "loop_control",
    "name",
    "no_log",
    "notify",
    "poll",
    "port",
    "register",
    "remote_user",
    "retries",
    "run_once",
    "tags",
    "until",
    "when",
]

TASKLIST_CACHE = {}
TASKLIST_EXTENSIONS = ["at.yaml", "at.yml"]


def find_tasklists_in_repos(tasklist_repos, exclude_paths=None):

    if exclude_paths is None:
        exclude_paths = []

    if isinstance(tasklist_repos, string_types):
        tasklist_repos = [tasklist_repos]

    result = {}
    tasklists_per_repo = {}

    for tlr in tasklist_repos:
        tasklists = find_tasklists_in_repo(tlr, exclude_paths=exclude_paths)
        # we want tasklists to be found first to take precedence
        for tln, content in tasklists.items():
            if tln not in result.keys():
                result[tln] = content
                tasklists_per_repo.setdefault(tlr, []).append(tln)

    return result, tasklists_per_repo


def find_tasklists_in_repo(tasklist_repo, exclude_paths=None):

    if exclude_paths is None:
        exclude_paths = []

    if tasklist_repo in TASKLIST_CACHE.keys():
        return TASKLIST_CACHE[tasklist_repo]

    result = {}
    try:
        for root, dirnames, filenames in os.walk(
            os.path.realpath(tasklist_repo), topdown=True, followlinks=True
        ):
            dirnames[:] = [d for d in dirnames if d not in DEFAULT_EXCLUDE_DIRS]
            dirnames[:] = [
                d for d in dirnames if os.path.join(root, d) not in exclude_paths
            ]
            # check for extensions

            for filename in filenames:
                match = False
                for ext in TASKLIST_EXTENSIONS:
                    if filename.endswith(".{}".format(ext)):
                        match = True
                        break

                if not match:
                    continue

                tl_file = os.path.join(root, filename)

                try:
                    format, tasklist = get_tasklist_file_format(tl_file)
                    if format != "ansible":
                        log.debug(
                            "Not ansible tasklist format, ignoring: {} -> {}".format(
                                tl_file, format
                            )
                        )
                        continue

                    # tl_name = os.path.splitext(filename)[0]
                    result[filename] = tasklist

                except (Exception) as e:
                    log.warning(
                        ("Could not parse tasklist '{}': {}".format(tl_file, e))
                    )
                    log.debug(e, exc_info=1)

    except (UnicodeDecodeError):
        print(
            " X one or more filenames under '{}' can't be decoded, ignoring. This can cause problems later. ".format(
                root
            )
        )

    TASKLIST_CACHE[tasklist_repo] = result
    # import pp
    # print("TASKLISTS")
    # pp(result)
    return result


def get_tasklist_file_format(path):

    if not isinstance(path, string_types):
        raise NsblException("tasklist file needs to be string: {}".format(path))

    tasklist = load_object_from_url_or_path(path)

    return get_tasklist_format(tasklist), tasklist


def get_tasklist_format(task_list):
    """This is a not quite 100% method to check whether a task list is in ansbile format, or freckle.
    """

    if not task_list or not isinstance(task_list, (list, tuple, CommentedSeq)):
        return "unknown"

    for item in task_list:

        if isinstance(item, string_types):
            log.debug(
                "task item '{}' is string, determining this is a 'freckles' task list".format(
                    item
                )
            )
            return "freckles"
        elif isinstance(item, dict):
            keys = set(item.keys())
            if keys & set(ANSIBLE_TASK_KEYWORDS):
                log.debug(
                    "task item keys ({}) contain at least one known Ansible keyword , determining this is 'ansible' task list format".format(
                        keys
                    )
                )
                return "ansible"
        else:
            raise Exception("Not a valid task-list item: {}".format(item))

    # TODO: log outupt
    # could check for 'meta' key above, but 'meta' can be a keyword in ansible too,
    # so figured I check for everything else first
    for item in task_list:
        if "frecklet" in item.keys():
            log.debug(
                "task item '{}' has 'frecklet' key, determining this is a 'freckles' task list".format(
                    item["frecklet"].get("name", item)
                )
            )
            return "freckles"
        for key in item.keys():
            if key.isupper():
                log.debug(
                    "task item key '{}' is all uppercase, determining this is a 'freckles' task list".format(
                        key
                    )
                )
                return "freckles"
    return "unknown"


def convert_ansible_tasklist(
    tasklist_name,
    tasklist,
    import_type="import",
    convert_ansible_template_markers_to_freckles=False,
):

    if import_type not in ["import", "include"]:
        raise Exception("Invalid import type: {}".format(import_type))

    if convert_ansible_template_markers_to_freckles:
        raise NotImplementedError(
            "Converting ansible template markers not implemented yet."
        )

    result = {
        "name": tasklist_name,
        # "command": tasklist_name,
        "import_type": import_type,
        "type": "ansible-tasklist",
        "tasklist_content": tasklist,
        "tasklist_var": "ansible_tasklist_{}".format(tasklist_name),
    }

    return result



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

import copy
import getpass
import io
import logging
import os
import re
import shutil
import tarfile
import threading
from collections import OrderedDict
from datetime import datetime
from distutils.spawn import find_executable
from random import randint
from time import sleep

import click
from cookiecutter.main import cookiecutter
from jinja2 import Environment, FileSystemLoader
from plumbum import local
from ruamel.yaml import YAML
from ruamel.yaml.comments import CommentedMap
from six import string_types

from frkl import load_string_from_url_or_path, FrklistContext
from frkl.exceptions import FrklistConfigException
from frkl.utils import get_url_parents
from frutils import can_passwordless_sudo, dict_merge, is_url_or_abbrev
from .defaults import (
    DEFAULT_ENV_TYPE,
    ADD_TYPE_TASK_LIST,
    ADD_TYPE_ROLE,
    ADD_TYPE_ACTION,
    ADD_TYPE_CALLBACK,
    ADD_TYPE_FILTER,
    ADD_TYPE_LIBRARY,
    ANSIBLE_ROLE_CACHE_DIR,
    ADD_TYPE_FILES,
    ADD_TYPE_MODULE_UTIL,
    ADD_TYPE_TASK_LIST_FILE,
    TEMPLATES_FOLDER,
    MITOGEN_VERSION,
)
from .exceptions import NsblException
from .inventory import NsblInventory
from .nsbl_tasklist import NsblTasklist

try:
    set
except NameError:
    from sets import Set as set

log = logging.getLogger("nsbl")

cookiecutter_lock = threading.Lock()


def create_yaml():
    yaml = YAML()
    yaml.default_flow_style = False
    yaml.preserve_quotes = True
    yaml.width = 4096
    return yaml


# class NsblContext(FrklistContext):
#
#     def __init__(self, **kwargs):
#
#         self.kwargs = kwargs


def create_single_host_nsbl_env_from_tasklist(
    task_list,
    context,
    user=None,
    host="localhost",
    host_ip=None,
    connection_type=None,
    ssh_port=22,
    ssh_key=None,
    python_interpreter=None,
    global_task_vars=None,
    pre_tasks=None,
    gather_facts=True,
    additional_files=None,
    task_marker="task",
    meta_marker="task",
    **kwargs
):

    if user is None:
        user = getpass.getuser()

    config_dict = [{host: {"meta": {"type": "host"}, "tasks": task_list}}]

    if connection_type is None:
        if host in ["localhost", "127.0.0.1"]:
            connection_type = "local"
        else:
            connection_type = "ssh"

    connection_vars = {
        "ansible_port": ssh_port,
        "ansible_connection": connection_type,
        "ansible_user": user,
    }
    if python_interpreter is not None:
        connection_vars["ansible_python_interpreter"] = python_interpreter
    if ssh_key is not None:
        connection_vars["ansible_ssh_private_key_file"] = ssh_key
    if host_ip is not None:
        connection_vars["ansible_host"] = host_ip

    config_dict[0][host]["vars"] = connection_vars

    config = Nsbl(
        [config_dict],
        context=context,
        base_path=None,
        global_task_vars=global_task_vars,
        pre_tasks=pre_tasks,
        gather_facts=gather_facts,
        additional_files=additional_files,
        task_marker=task_marker,
        meta_marker=meta_marker,
    )

    return config


def create_nsbl_env(
    urls,
    base_path=None,
    context=None,
    default_env_type=DEFAULT_ENV_TYPE,
    additional_files=None,
    global_task_vars=None,
    task_marker="task",
    meta_marker="task",
):

    if isinstance(urls, string_types):
        urls = [urls]  # we always want a list of lists as input for the Nsbl object

    if is_url_or_abbrev(urls[0]):
        click.echo("Downloading configuration: {}".format(urls[0]))
    config_dicts = load_string_from_url_or_path(
        urls, create_python_object=True, safe_load=False
    )

    if not base_path:
        parent = get_url_parents(urls, return_list=True)
        if len(parent) > 2:
            raise NsblException(
                "Multiple base paths calculated ({}). This is not allowed as it introduces unpredictable behaviour. Please specify the base path manually.".format(
                    parent
                )
            )
        base_path = list(parent)[0]

    config = Nsbl(
        config_dicts,
        base_path=base_path,
        context=context,
        default_env_type=default_env_type,
        additional_files=additional_files,
        global_task_vars=global_task_vars,
        task_marker=task_marker,
        meta_marker=meta_marker,
    )

    return config


# def expand_nsbl_config(configs):
#     """Expands the nsbl configuration.
#
#     Args:
#         configs (list): a list of configuration items
#     """
#
#     chain = [FrklProcessor(NSBL_INVENTORY_BOOTSTRAP_FORMAT)]
#     f = Frkl(configs, chain)
#     result = f.process()
#
#     return result

DEFAULT_ELEVATED_PERMISSIONS_REQUIRED = True
DEFAULT_PASSWORDLESS_SUDO_POSSIBLE = False


class Nsbl(object):
    """Holds and parses configuration to generate Ansible task lists and inventories.

    Args:
        config (list): a list of configuration items
        base_path (str): the base path to use for relative task list imports etc.
        context (NsblContext): the context for this environment
        default_env_type (str): the type a environment is if it is not explicitely specified, either ENV_TYPE_HOST or ENV_TYPE_GROUP
        additional_files (dict): a dict of additional files to copy into the Ansible environment
        # allow_external_roles (bool): whether to allow the downloading of external roles
        global_task_vars (dict): global variables for all tasks (will still be overwritten by task-list specific task vars if present)
        pre_tasks (list): a list of tasks that will be inserted into every playbook
        gather_facts (boolean): whether the gather_facts step should be enabled/disabled
        task_marker (string): the marker string that indicates a task metadata key
        meta_marker (string): the maker string that indicates a 'meta' metadata key (for getting the description of a task)
    """

    def __init__(
        self,
        config,
        base_path=None,
        context=None,
        default_env_type=DEFAULT_ENV_TYPE,
        additional_files=None,
        global_task_vars=None,
        pre_tasks=None,
        gather_facts=True,
        task_marker="task",
        meta_marker="task",
    ):

        self.base_path = base_path
        self.plays = CommentedMap()
        self.internal_roles = set()
        self.external_roles = set()

        self.task_marker = task_marker
        self.meta_marker = meta_marker

        self.pre_tasks = pre_tasks
        self.gather_facts = gather_facts

        if global_task_vars is None:
            global_task_vars = {}
        self.global_task_vars = global_task_vars

        self.config = config
        self._lock = threading.Lock()
        if context is None:
            context = FrklistContext()
        elif isinstance(context, (dict, CommentedMap, OrderedDict)):
            context = FrklistContext(**context)
        elif not isinstance(context, FrklistContext):
            raise FrklistConfigException(
                "Invalid type for context: {}".format(type(context))
            )
        self.context = context

        self.default_env_type = default_env_type

        if additional_files is None:
            additional_files = {}
        self.additional_files = copy.deepcopy(additional_files)

        self.inventory = NsblInventory.create(
            self.config, default_env_type=self.default_env_type, pre_chain=[]
        )

        tl_id = 0
        for tasks in self.inventory.tasks:

            task_list_meta = tasks["meta"]
            env_name = task_list_meta["name"]
            env_id = task_list_meta["_env_id"]

            # task_list_vars = tasks.get("vars", {})
            # print(task_list_vars)
            # dict_merge(task_list_vars, self.global_task_vars, copy_dct=False)
            task_list_vars = self.global_task_vars

            task_list = tasks["tasks"]

            meta = {}
            meta["tasklist_id"] = tl_id
            meta["env_id"] = env_id
            meta["env_name"] = env_name
            # meta["tasklist_parent"] = self.base_path

            tl = NsblTasklist(
                task_list,
                context=self.context,
                meta=meta,
                vars=task_list_vars,
                task_marker=self.task_marker,
                meta_marker=self.meta_marker,
            )
            self.plays["{}_{}".format(env_name, env_id)] = {
                "task_list": tl,
                "meta": task_list_meta,
                "env_id": env_id,
                "env_name": env_name,
            }

            tl_id = tl_id + 1
            self.internal_roles.update(tl.internal_roles)
            self.external_roles.update(tl.external_roles)

    def get_play(self, env_name, env_id):

        return self.plays.get("{}_{}".format(env_name, env_id), None)

    def get_tasklist(self, env_name, env_id):

        return self.get_play(env_name, env_id)["task_list"]

    def render(
        self,
        env_dir,
        global_vars=None,
        extract_vars=True,
        force=False,
        elevated_permissions_required=None,
        passwordless_sudo_possible=None,
        secure_vars=None,
        ansible_args="",
        callback="default",
        force_update_roles=False,
        add_timestamp_to_env=False,
        add_symlink_to_env=False,
        extra_paths=None,
        show_tasks_with_password_in_log=False,
        use_mitogen=False,
        use_ara=False,
        parent_task=None,
    ):
        """Creates the ansible environment in the folder provided.

        Args:
          env_dir (str): the folder where the environment should be created
          global_vars (dict): vars to be rendered as global on top of a playbook
          extract_vars (bool): whether to extract a hostvars and groupvars directory for the inventory (True), or render a dynamic inventory script for the environment (default, True) -- Not supported at the moment
          force (bool): overwrite environment if already present at the specified location, use with caution because this might delete an important folder if you get the 'target' dir wrong
          elevated_permissions_required (bool): whether to include the '--ask-become-pass' arg to the ansible-playbook call
          passwordless_sudo_possible (bool): whether the user executing this run does have passwordless sudo permissions
          secure_vars (dict): vars to keep in a vault (not implemented yet)
          ansible_args (str): parameters to give to ansible-playbook (like: "-vvv")
          callback (str): name of the callback to use, default: nsbl_internal
          force_update_roles (bool): whether to overwrite external roles that were already downloaded
          add_timestamp_to_env (bool): whether to add a timestamp to the env_dir -- useful for when this is called from other programs (e.g. freckles)
          add_symlink_to_env (str): whether to add a symlink to the current env from a fixed location (useful to archive all runs/logs)
          extra_paths (str): a list of of extra paths to be exported before the ansible playbook run
          show_tasks_with_password_in_log (bool): whether to show tasks that contain a password variable in the log (only enable for debugging!)
          use_mitogen (bool): use mitogen as a strategy plugin to speed up execution
          use_ara (bool): use ara (https://ara.readthedocs.io)
          parent_task (TaskDetail): parent task (for task callbacks, optional)
        """
        if elevated_permissions_required is None:
            elevated_permissions_required = DEFAULT_ELEVATED_PERMISSIONS_REQUIRED

        if not isinstance(elevated_permissions_required, bool):
            raise Exception("'ask_become_pass' needs to be boolean value")

        if extra_paths is None:
            extra_paths = []

        if passwordless_sudo_possible is None:
            if self.inventory.localhost_only:
                passwordless_sudo_possible = can_passwordless_sudo()
            else:
                passwordless_sudo_possible = DEFAULT_PASSWORDLESS_SUDO_POSSIBLE

        env_dir = os.path.expanduser(env_dir)
        if add_timestamp_to_env:
            start_date = datetime.now()
            date_string = start_date.strftime("%y%m%d_%H_%M_%S")
            dirname, basename = os.path.split(env_dir)
            env_dir = os.path.join(dirname, "{}_{}".format(basename, date_string))

        if global_vars is None:
            global_vars = {}

        result = {}
        result["env_dir"] = env_dir
        if os.path.exists(env_dir) and force:
            shutil.rmtree(env_dir)

        inventory_dir = os.path.join(env_dir, "inventory")
        result["inventory_dir"] = inventory_dir

        if extract_vars:
            inv_target = "../inventory/hosts"
        else:
            inv_target = "../inventory/inventory"

        result["extract_vars"] = extract_vars

        playbook_dir = os.path.join(env_dir, "plays")
        result["playbook_dir"] = playbook_dir
        # roles_base_dir = os.path.join(env_dir, "roles")
        result["roles_base_dir"] = playbook_dir

        # if password is None:
        #     if elevated_permissions_required:
        #         if passwordless_sudo_possible:
        #             ask_sudo = False
        #         else:
        #             ask_sudo = True
        #     else:
        #         ask_sudo = False
        # else:
        #     ask_sudo = True

        all_plays_name = "all_plays.yml"
        result["default_playbook_name"] = all_plays_name

        ansible_playbook_args = ansible_args
        result["ansible_playbook_cli_args"] = ansible_playbook_args
        result["run_playbooks_script"] = os.path.join(env_dir, "run_all_plays.sh")

        ara_base = None
        if use_ara:

            ara_success = False

            python_exe = find_executable("python", path=extra_paths)

            if python_exe:
                try:
                    ara_base = local[python_exe](["-m", "ara.setup.path"])
                    if ara_base:
                        ara_base = ara_base.strip()
                        ara_success = True
                except (Exception):
                    pass

            if not ara_success:
                ara_base = None
                log.warning("Did not find 'ara' in python env, not using it...")

        if ara_base:
            callback_plugins_list = "callback_plugins:{}/plugins/callback".format(
                ara_base
            )
            action_plugins_list = "action_plugins:{}/plugins/action".format(ara_base)
            # library_plugins_list = "library:{}/plugins/modules".format(ara_base)
            # module_utils_list = "module_utils:{}/plugins/module_utils".format(ara_base)
        else:
            callback_plugins_list = "callback_plugins"
            action_plugins_list = "action_plugins"

        library_plugins_list = "library"
        module_utils_list = "module_utils"

        if not use_mitogen:
            mitogen = ""
        else:
            mitogen_path = "../opt/mitogen-{}".format(MITOGEN_VERSION)
            mitogen = os.path.join(
                mitogen_path, "ansible_mitogen", "plugins", "strategy"
            )

        cookiecutter_details = {
            "inventory": inv_target,
            "env_dir": env_dir,
            "extra_paths": extra_paths,
            "playbook_dir": playbook_dir,
            "ansible_playbook_args": ansible_playbook_args,
            "library_path": library_plugins_list,
            "action_plugins_path": action_plugins_list,
            "module_utils_path": module_utils_list,
            "extra_script_commands": "",
            "ask_sudo": "",
            "playbook": all_plays_name,
            "callback_plugins": callback_plugins_list,
            "callback_plugin_name": callback,
            "callback_whitelist": "default_to_file",
            "mitogen": mitogen,
            "use_ara": "true" if use_ara else "false",
        }

        log.debug("Creating build environment from template...")
        log.debug("Using cookiecutter details: {}".format(cookiecutter_details))

        template_path = os.path.join(
            os.path.dirname(__file__), "external", "nsbl-environment-template"
        )

        create_cookiecutter_env(
            template_path=template_path, cookiecutter_details=cookiecutter_details
        )

        if add_symlink_to_env:
            link_path = os.path.expanduser(add_symlink_to_env)
            if os.path.exists(link_path) or os.path.islink(link_path):
                os.unlink(link_path)
            link_parent = os.path.abspath(os.path.join(link_path, os.pardir))
            try:
                os.makedirs(link_parent)
            except (Exception):
                pass

            os.symlink(env_dir, link_path)

            result["env_dir_link"] = link_path

        # write inventory
        if extract_vars:
            self.inventory.extract_vars(inventory_dir)

        self.inventory.write_inventory_file_or_script(
            inventory_dir, extract_vars=extract_vars
        )

        # extract mitogen
        if use_mitogen:
            mitogen_archive_path = os.path.join(
                os.path.dirname(__file__),
                "external",
                "mitogen",
                "mitogen-{}.tar.gz".format(MITOGEN_VERSION),
            )
            opt_path = os.path.join(env_dir, "opt")
            if not os.path.exists(opt_path):
                os.makedirs(opt_path)
            mitogen_archive = tarfile.open(mitogen_archive_path)
            mitogen_archive.extractall(path=opt_path)
            mitogen_archive.close()

        # write roles
        all_playbooks = []
        all_playbook_names = []
        ext_roles = []
        task_details = []

        for play, task_list_details in self.plays.items():

            # td = copy.deepcopy(task_list_details)
            tasks = task_list_details["task_list"]
            id = task_list_details["env_id"]
            name = task_list_details["env_name"]
            playbook_name = "play_{}_{}.yml".format(name, id)
            playbook_file = os.path.join(playbook_dir, playbook_name)

            playbook_vars = {}
            dict_merge(playbook_vars, global_vars, copy_dct=False)
            dict_merge(playbook_vars, tasks.vars, copy_dct=False)
            playbook_vars["_env_id"] = id
            playbook_vars["_env_name"] = name

            task_list = tasks.render_tasklist(
                secure_vars=secure_vars,
                show_tasks_with_password_in_log=show_tasks_with_password_in_log,
            )

            if self.pre_tasks:
                # TODO: clean that up, tasks could now go directly into the playbook
                playbook_pre_file = os.path.join(
                    playbook_dir, "pre_{}".format(playbook_name)
                )
                tasklist_name = "pre_tasks_{}_{}".format(name, id)
                tasklist_name = re.sub(r"\W+", "", tasklist_name)
                tl_pre = {
                    "type": ADD_TYPE_TASK_LIST,
                    "target_name": "{}.yml".format(tasklist_name),
                    "tasklist": self.pre_tasks,
                    "var_name": tasklist_name,
                }
                self.additional_files[tasklist_name] = tl_pre
                import_pre_tasks = {
                    "name": "[importing host preparation tasks]",
                    "import_tasks": "{{{{ {} }}}}".format(tasklist_name),
                }
                playbook_dict_pre = CommentedMap()
                playbook_dict_pre["hosts"] = name
                playbook_dict_pre["strategy"] = "linear"
                playbook_dict_pre["gather_facts"] = False
                playbook_dict_pre["vars"] = playbook_vars
                playbook_dict_pre["tasks"] = [import_pre_tasks]
                all_playbooks.append(
                    {
                        "name": "pre_{}".format(playbook_name),
                        "dict": playbook_dict_pre,
                        "file": playbook_pre_file,
                    }
                )
                all_playbook_names.append("pre_{}".format(playbook_name))

            playbook_dict = CommentedMap()
            playbook_dict["hosts"] = name

            playbook_dict["gather_facts"] = self.gather_facts

            playbook_dict["vars"] = playbook_vars
            playbook_dict["tasks"] = task_list

            # td["playbook"] = playbook_dict
            all_playbooks.append(
                {"name": playbook_name, "dict": playbook_dict, "file": playbook_file}
            )
            all_playbook_names.append(playbook_name)

            dict_merge(self.additional_files, tasks.additional_files, copy_dct=False)

            if tasks.external_roles:
                for n in tasks.external_roles:
                    if n not in ext_roles:
                        ext_roles.append(n)

        # copy external files
        external_files_vars = {}

        action_plugins_target = os.path.join("plays", "action_plugins")
        callback_plugins_target = os.path.join("plays", "callback_plugins")
        filter_plugins_target = os.path.join("plays", "filter_plugins")
        library_plugins_target = os.path.join("plays", "library")
        module_utils_target = os.path.join("plays", "module_utils")
        roles_target = os.path.join("roles", "internal")
        task_lists_target = "task_lists"
        files_target = "files"

        additional_files = copy.deepcopy(self.additional_files)

        for path, details in additional_files.items():

            file_type = details["type"]

            # if not os.path.exists(path):
            #     import pp
            #     print(path)
            #     pp(details)
            #     raise NsblException(
            #         "Resource '{}' not available: {}".format(file_type, path)
            #     )

            playbook_var_name = details.get("var_name", None)
            file_name = details["target_name"]

            skip_copy = False
            if file_type == ADD_TYPE_TASK_LIST:
                # content = details["tasklist"]
                target = os.path.join(env_dir, task_lists_target, file_name)
                log.debug("Render 'ansible'-type tasklist: {}".format(path))

                target_parent = os.path.dirname(target)
                if not os.path.exists(target_parent):
                    os.makedirs(target_parent)

                with io.open(target, "w", encoding="utf-8") as tlf:
                    yaml = create_yaml()
                    yaml.dump(details["tasklist"], tlf)

                skip_copy = True
                playbook_var_value = os.path.join(
                    "{{ playbook_dir }}", "..", task_lists_target, file_name
                )

            elif file_type == ADD_TYPE_ROLE:
                target = os.path.join(env_dir, roles_target, file_name)
                playbook_var_value = None
                copy_source_type = "dir"
            elif file_type == ADD_TYPE_ACTION:
                target = os.path.join(env_dir, action_plugins_target, file_name)
                playbook_var_value = None
                copy_source_type = "file"
            elif file_type == ADD_TYPE_CALLBACK:
                target = os.path.join(env_dir, callback_plugins_target, file_name)
                playbook_var_value = None
                copy_source_type = "file"
            elif file_type == ADD_TYPE_FILTER:
                target = os.path.join(env_dir, filter_plugins_target, file_name)
                playbook_var_value = None
                copy_source_type = "file"
            elif file_type == ADD_TYPE_LIBRARY:
                target = os.path.join(env_dir, library_plugins_target, file_name)
                playbook_var_value = None
                copy_source_type = "file"
            elif file_type == ADD_TYPE_MODULE_UTIL:
                target = os.path.join(env_dir, module_utils_target, file_name)
                copy_source_type = "file"
                playbook_var_value = None
            elif file_type == ADD_TYPE_FILES:
                target = os.path.join(env_dir, files_target, file_name)
                copy_source_type = "file"
                playbook_var_value = None
            elif file_type == ADD_TYPE_TASK_LIST_FILE:
                target = os.path.join(env_dir, task_lists_target, file_name)
                copy_source_type = "file"
                playbook_var_value = None
            else:
                raise NsblException("Invalid external file type: {}".format(file_type))

            if not skip_copy:
                log.debug("Copying {} '{}': {}".format(file_type, file_name, target))
                target_parent = os.path.dirname(target)
                if not os.path.exists(target_parent):
                    os.makedirs(target_parent)
                if not os.path.isdir(os.path.realpath(target_parent)):
                    raise NsblException(
                        "Can't copy files to '{}': not a directory".format(
                            target_parent
                        )
                    )

                if copy_source_type == "file":
                    shutil.copyfile(path, target)
                else:
                    shutil.copytree(path, target)

            if playbook_var_value:
                if playbook_var_name in external_files_vars.keys():
                    raise NsblException(
                        "Duplicate key for external files: {}".format(playbook_var_name)
                    )
                log.debug(
                    "Setting variable '{}' to: {}".format(
                        playbook_var_name, playbook_var_value
                    )
                )
                external_files_vars[playbook_var_name] = playbook_var_value

        # render all playbooks
        for playbook in all_playbooks:

            playbook_file = playbook["file"]
            playbook_dict = playbook["dict"]

            # adding external files vars
            dict_merge(
                playbook_dict.setdefault("vars", {}),
                external_files_vars,
                copy_dct=False,
            )

            with io.open(playbook_file, "w", encoding="utf-8") as pf:
                yaml = create_yaml()
                yaml.dump([playbook_dict], pf)

        result["task_details"] = task_details
        result["additional_files"] = self.additional_files
        jinja_env = Environment(loader=FileSystemLoader(TEMPLATES_FOLDER))
        template = jinja_env.get_template("play.yml")
        output_text = template.render(playbooks=all_playbook_names)

        all_plays_file = os.path.join(env_dir, "plays", all_plays_name)
        result["all_plays_file"] = all_plays_file
        with io.open(all_plays_file, "w", encoding="utf-8") as text_file:
            text_file.write(output_text)

        if ext_roles:

            ext_roles_target = os.path.join(env_dir, "roles", "external")
            # render roles_requirements.yml
            jinja_env = Environment(loader=FileSystemLoader(TEMPLATES_FOLDER))
            roles_requirements_file = os.path.join(
                ext_roles_target, "roles_requirements.yml"
            )

            if not os.path.exists(ext_roles_target):
                os.makedirs(ext_roles_target)

            roles_to_copy = {}
            for role in ext_roles:
                role_src = os.path.join(ANSIBLE_ROLE_CACHE_DIR, role)
                target = os.path.join(ext_roles_target, role)
                roles_to_copy[role_src] = target

                template = jinja_env.get_template("external_role.yml")
                output_text = template.render(role={"src": role, "name": role})
                with io.open(roles_requirements_file, "a", encoding="utf-8") as myfile:
                    myfile.write(output_text)

            # download external roles
            role_requirement_file = os.path.join(
                env_dir, "roles", "external", "roles_requirements.yml"
            )

            if not os.path.exists(ANSIBLE_ROLE_CACHE_DIR):
                os.makedirs(ANSIBLE_ROLE_CACHE_DIR)

            command = [
                "install",
                "-r",
                role_requirement_file,
                "-p",
                ANSIBLE_ROLE_CACHE_DIR,
            ]
            if force_update_roles:
                command.append("--force")
            log.debug("Downloading and installing external roles...")

            if parent_task:
                download_task = parent_task.add_subtask(
                    task_name="download external roles",
                    msg="downloading role(s): {}".format(" ".join(ext_roles)),
                )
            else:
                download_task = None
                click.echo("Managing external roles...")

            # path = os.environ["PATH"] + ":" + extra_paths
            path = extra_paths
            galaxy_exe = find_executable("ansible-galaxy", path=path)
            if not galaxy_exe:
                raise Exception("Could not find 'ansible-galaxy' executable")
            ansible_galaxy = local[galaxy_exe]

            try:
                rc, stdout, stderr = ansible_galaxy.run(command)

                if download_task is None:
                    for line in stdout.split("\n"):
                        if (
                            "already installed" not in line
                            and "--force to change" not in line
                            and "unspecified" not in line
                        ):
                            # log.debug("Installing role: {}".format(line.encode('utf8')))

                            click.echo("  {}".format(line.encode("utf8")), nl=False)
                else:
                    if rc == 0:
                        success = True
                    else:
                        success = False

                    download_task.finish(
                        success=success,
                        msg=stdout,
                        error_msg=stderr,
                        skipped=False,
                        changed=True,
                    )

            except (Exception) as e:

                if download_task is not None:
                    download_task.finish(
                        success=False, error_msg=str(e), skipped=False, changed=True
                    )
                raise e

            if roles_to_copy:
                if parent_task is not None:
                    copy_task = parent_task.add_subtask(
                        task_name="copy roles",
                        msg="copying role(s) into Ansible environment",
                    )
                else:
                    copy_task = None
                    if len(ext_roles) == 1:

                        click.echo(
                            "Copying role from Ansible cache: {}".format(ext_roles[0])
                        )
                    else:
                        click.echo("Copying roles from Ansible cache:")
                        for r in ext_roles:
                            click.echo("  - {}".format(r))

                for src, target in roles_to_copy.items():
                    log.debug("Coping external role: {} -> {}".format(src, target))
                    shutil.copytree(src, target)

                if copy_task is not None:
                    copy_task.finish(success=True, skipped=False, changed=True)

        return result


def create_cookiecutter_env(
    template_path, cookiecutter_details, overwrite_if_exists=False
):

    with cookiecutter_lock:
        try:
            cookiecutter(
                template_path,
                extra_context=cookiecutter_details,
                no_input=True,
                overwrite_if_exists=overwrite_if_exists,
            )
        except (Exception) as e:
            # print("XXXXXXXXXXXXXXXXXXXXXXXXXX: {}".format(e))
            # this is weird, sometimes when trying to run parallel freckles job, this fails

            i = 0
            success = False
            last_exception = e
            while i < 6:
                i = i + 1
                sl = randint(500, 1500) / 1000
                log.debug(
                    "Error when trying to create Ansible env dir, retrying after sleeping {} milliseconds: {}".format(
                        sl, e
                    )
                )
                sleep(sl)
                try:
                    cookiecutter(
                        template_path,
                        extra_context=cookiecutter_details,
                        no_input=True,
                        overwrite_if_exists=True,
                    )
                    success = True
                    break
                except (Exception) as le:
                    # print("yyyyyyyyyyyyyyy {}".format(le))
                    last_exception = le
            if not success:
                raise last_exception



# -*- coding: utf-8 -*-

"""Top-level package for nsbl."""
import io
import os

from pkg_resources import get_distribution, DistributionNotFound

__author__ = """Markus Binsteiner"""
__email__ = "makkus@frkl.io"

try:
    # Change here if project is renamed and does not equal the package name
    dist_name = __name__
    __version__ = get_distribution(dist_name).version
except DistributionNotFound:
    try:
        version_file = os.path.join(os.path.dirname(__file__), "version.txt")

        if os.path.exists(version_file):
            with io.open(version_file, encoding="utf-8") as vf:
                __version__ = vf.read()
        else:
            __version__ = "unknown"

    except (Exception):
        pass

    if __version__ is None:
        __version__ = "unknown"
finally:
    del get_distribution, DistributionNotFound



# -*- coding: utf-8 -*-

# python 3 compatibility
from __future__ import absolute_import, division, print_function, unicode_literals

import copy
import logging
import os
from collections import Mapping

from ruamel.yaml.comments import CommentedMap
from six import string_types

from frkl import Frklist, FrklistContext, load_object_from_url_or_path
from frutils import StringYAML, is_url_or_abbrev
from frutils.exceptions import FrklException
from frutils.frutils import valid_python_var_name
from .defaults import (
    ADD_TYPE_ROLE,
    ADD_TYPE_TASK_LIST,
    DEFAULT_INCLUDE_TYPE,
    FRECKLES_DESC_METADATA_KEY,
    FRECKLES_DESC_SHORT_METADATA_KEY,
)
from .exceptions import NsblException
from .role_utils import find_roles_in_repos
from .tasklist_utils import ANSIBLE_TASK_KEYWORDS, find_tasklists_in_repos

GLOBAL_ENV_ID_COUNTER = 1110
GLOBAL_TASKLIST_ID_COUNTER = 1110


def GLOBAL_TASKLIST_ID():
    global GLOBAL_TASKLIST_ID_COUNTER
    GLOBAL_TASKLIST_ID_COUNTER = GLOBAL_TASKLIST_ID_COUNTER + 1
    return GLOBAL_TASKLIST_ID_COUNTER


def GLOBAL_ENV_ID():
    global GLOBAL_ENV_ID_COUNTER
    GLOBAL_ENV_ID_COUNTER = GLOBAL_ENV_ID_COUNTER + 1
    return GLOBAL_ENV_ID_COUNTER


yaml = StringYAML()
yaml.default_flow_style = False

log = logging.getLogger("nsbl")


class NsblContext(FrklistContext):
    def __init__(
        self, allow_external_roles=None, allow_external_tasklists=None, **kwargs
    ):
        super(NsblContext, self).__init__(**kwargs)

        if allow_external_roles is None:
            allow_external_roles = self.allow_remote
        self.allow_external_roles = allow_external_roles

        if allow_external_tasklists is None:
            allow_external_tasklists = self.allow_remote
        self.allow_external_tasklists = allow_external_tasklists

        self.role_repo_paths = self.urls.get("roles", [])
        self.tasklist_paths = self.urls.get("tasklists", [])

        self.available_roles, all_role_paths = find_roles_in_repos(self.role_repo_paths)
        self.available_tasklists, self.tasklists_per_repo = find_tasklists_in_repos(
            self.tasklist_paths, exclude_paths=all_role_paths
        )  # we don't want tasklists from any roles

    def get_role_path(self, role_name):

        if role_name in self.available_roles.keys():
            return self.available_roles[role_name]
        else:
            if self.allow_external_roles:
                return role_name
            else:
                raise Exception(
                    "External roles not allowed, and role '{}' not in role paths: {}".format(
                        role_name, self.role_repo_paths
                    )
                )

    def get_tasklist(self, tasklist_name):

        if tasklist_name in self.available_tasklists.keys():
            return self.available_tasklists[tasklist_name]
        else:

            if os.path.exists(tasklist_name):
                return load_object_from_url_or_path(tasklist_name)
            elif is_url_or_abbrev(tasklist_name):
                if not self.allow_external_tasklists or self.allow_remote:
                    raise Exception(
                        "Remote tasklist not allowed, not retrieving: {}".format(
                            tasklist_name
                        )
                    )
                try:
                    return load_object_from_url_or_path(tasklist_name)
                except (Exception) as e:
                    raise Exception(
                        "Could not download tasklist '{}': {}".format(tasklist_name, e)
                    )
            else:
                raise Exception(
                    "Could not find Ansible tasklist '{}' in tasklist paths: {}".format(
                        tasklist_name, self.tasklist_paths
                    )
                )

    def get_repo_for_tasklist(self, tasklist_name):

        for repo, tlns in self.tasklists_per_repo.items():

            if tasklist_name in tlns:
                return repo

        return None


class NsblTasklist(Frklist):
    def __init__(
        self,
        itemlist,
        context,
        meta=None,
        vars=None,
        task_marker="task",
        meta_marker="task",
    ):

        if not isinstance(context, NsblContext):
            raise Exception("Context needs to be of type NsblContext")

        self.internal_roles = set()
        self.external_roles = set()
        self.modules_used = set()
        self.tasklist_files = {}
        self.children = []
        self.additional_files = {}

        self.task_lookup = {}

        self.task_marker = task_marker
        self.meta_marker = meta_marker

        super(NsblTasklist, self).__init__(itemlist, context, meta=meta, vars=vars)

    def expand_and_augment_tasklist(self, tasklist):

        result = []
        for task in tasklist:

            index = task[self.meta_marker]["_task_id"]

            self.task_lookup[index] = task

            name = task[self.meta_marker].get("name", None)
            command = task[self.task_marker].get("command", None)
            if name is None:
                if command is None:
                    raise Exception(
                        "Task doesn't have 'name' nor 'command' key: {}".format(task)
                    )
                name = command
                task[self.meta_marker]["name"] = command
            if command is None:
                command = name
                task[self.task_marker]["command"] = command
            temp = {
                self.task_marker: copy.deepcopy(task[self.task_marker]),
                "vars": task.get("vars", {}),
            }

            task_type = task[self.meta_marker]["type"]
            temp[self.task_marker]["type"] = task_type

            desc_val = task.get(self.meta_marker, {}).get(
                FRECKLES_DESC_METADATA_KEY, {}
            )
            msg = desc_val.get(
                FRECKLES_DESC_SHORT_METADATA_KEY, "{}: {}".format(task_type, command)
            )

            if msg is None:
                msg = task.get("doc", {}).get("msg", None)

            if msg is not None:
                temp[self.task_marker]["name"] = "[{}] {}".format(index, msg)
            else:
                temp[self.task_marker]["name"] = "[{}] {}".format(index, name)

            if temp[self.task_marker]["type"] == "ansible-role":

                if "include-type" not in temp[self.task_marker].keys():
                    temp[self.task_marker]["include-type"] = DEFAULT_INCLUDE_TYPE

                temp.setdefault("vars", {})["_task_id"] = index

                if "role_path" in temp[self.task_marker].keys():
                    path = temp[self.task_marker]["role_path"]
                else:
                    path = self.context.get_role_path(command)
                    temp[self.task_marker]["role_path"] = path

                self.add_role_to_addition_files(command, path=path)

            if temp[self.task_marker]["type"] == "ansible-tasklist":
                include_type = temp[self.task_marker].get("include-type", None)
                if include_type is None:
                    include_type = DEFAULT_INCLUDE_TYPE
                    temp[self.task_marker]["include-type"] = include_type

                temp.setdefault("vars", {})["_task_id"] = index

                tasklist_var = temp[self.task_marker].get("tasklist_var", None)
                if tasklist_var is None:
                    tasklist_var = "tasklist_{}".format(command)

                tasklist_var = valid_python_var_name(tasklist_var)
                temp[self.task_marker]["tasklist_var"] = tasklist_var

                self.add_tasklist_to_additional_files(command, tasklist_var)

            # resources
            resources = task[self.meta_marker].get("resources", {})

            for res_type, res_urls in resources.items():
                if res_type == "ansible-module":
                    # no need to do anything, as those are included
                    continue
                if res_type not in ["ansible-role", "ansible-tasklist"]:
                    continue

                if res_type == "ansible-role":
                    roles = res_urls
                    if isinstance(roles, string_types):
                        roles = [roles]

                    for role in roles:
                        self.add_role_to_addition_files(role, path=None)

                elif res_type == "ansible-tasklist":

                    tasklists = res_urls
                    if isinstance(tasklists, string_types):
                        tasklists = [tasklists]

                    for tl_name in tasklists:

                        tasklist_var = "tasklist_{}".format(tl_name)

                        tasklist_var = valid_python_var_name(tasklist_var)
                        task[self.task_marker]["tasklist_var"] = tasklist_var
                        self.add_tasklist_to_additional_files(tl_name, tasklist_var)

            result.append(temp)

            # moving 'register' value to 'task' key
            reg_val = task[self.meta_marker].get("register", None)
            if not reg_val:
                continue
            if not isinstance(reg_val, Mapping):
                raise TypeError(
                    "'register' value must be a mapping type: {}".format(reg_val)
                )
            temp.setdefault(self.task_marker, {})["register"] = reg_val

            # import pp, sys
            # pp(reg_val)
            # sys.exit()
            #
            # if (
            #     reg_val is not None
            #     and isinstance(reg_val, string_types)
            #     or (isinstance(reg_val, Mapping) and "var" in reg_val.keys())
            # ):
            #
            #     register_val = task[self.meta_marker]["register"]
            #     if isinstance(register_val, string_types):
            #         register_name = register_val
            #         register_query = None
            #     elif isinstance(register_val, Mapping):
            #         if "var" in register_val.keys():
            #
            #             register_name = register_val["var"]
            #             register_query = register_val.get("query", None)
            #
            #     if (
            #         "register" in temp[self.task_marker].keys()
            #         and temp[self.task_marker]["register"] != register_name
            #     ):
            #         t = copy.deepcopy(task)
            #         t[self.meta_marker].pop("secret_vars", None)
            #         t[self.meta_marker].pop("_task_id", None)
            #         raise FrklException(
            #             msg="Different 'register' values specified in both '{}' and '{}' subkeys".format(
            #                 self.task_marker, self.meta_marker
            #             ),
            #             solution="Check your task description for this task:\n\n{}\n\nEither remove one of the 'register' keys, or make sure they both have the same value.".format(
            #                 readable_yaml(t, indent=4)
            #             ),
            #         )
            #
            #     rn = register_name
            #     # TODO: check for other special characters
            #     if "-" in rn:
            #         t = copy.deepcopy(task)
            #         t[self.meta_marker].pop("secret_vars", None)
            #         t[self.meta_marker].pop("_task_id", None)
            #         raise FrklException(
            #             msg="Invalid 'register' keyname '{}', must be a valid variable name (can't contain '-', special charcters...)".format(
            #                 rn
            #             ),
            #             solution="Check your task description for this task:\n\n{}\n\nChange the 'register' variable name ('{}') to not contain any special characters.".format(
            #                 readable_yaml(t, indent=4), rn
            #             ),
            #         )
            #
            #     temp[self.task_marker]["register"] = rn
            #     if register_query is not None:
            #         temp[self.task_marker]["register_query"] = register_query

        return result

    def get_task(self, index):

        return self.task_lookup.get(index, None)

    def render_tasklist(self, secure_vars=None, show_tasks_with_password_in_log=False):
        """Renders the playbook into a file."""

        result = []

        for t in self.itemlist:

            task = copy.deepcopy(t)
            log.debug("Task item: {}".format(task))
            name = task[self.task_marker].pop("name")
            command = task[self.task_marker].pop("command", None)
            task_type = task[self.task_marker].pop("type")
            desc_val = task[self.task_marker].pop(FRECKLES_DESC_METADATA_KEY, {})
            desc = desc_val.get(FRECKLES_DESC_SHORT_METADATA_KEY, None)
            # legacy
            task_desc = task[self.task_marker].pop("task-desc", None)
            if desc is None:
                desc = task_desc
            if desc is None:
                desc = name

            vars = CommentedMap()
            task_has_password = False

            for key, value in task.pop("vars").items():

                # for alias in secure_vars.keys():
                #
                #     pw_wrap_method = secure_vars[alias]["type"]
                #     if pw_wrap_method != "environment":
                #         raise NsblException(
                #             "Password-wrap method '{}' not supported in nsbl.".format(
                #                 pw_wrap_method
                #             )
                #         )
                #     value, changed = simple_replace_string_in_obj(
                #         value, alias, "{{{{ lookup('env', '{}') }}}}".format(alias)
                #     )
                #     if changed:
                #         task_has_password = True

                vars[key] = value

            task_item = CommentedMap()
            task_item["name"] = desc

            if task_type == "ansible-module":
                if command is None:
                    raise NsblException(
                        "No 'command' key specified in task: {}".format(task)
                    )
                if "free_form" in vars.keys():
                    temp = copy.deepcopy(vars)
                    free_form = temp.pop("free_form")
                    task_item[command] = free_form
                    task_item["args"] = temp
                else:
                    task_item[command] = vars
            elif task_type == "ansible-tasklist":
                include_type = task[self.task_marker]["include-type"]
                task_key = "{}_tasks".format(include_type)
                task_item[task_key] = "{{{{ {} }}}}".format(
                    task[self.task_marker]["tasklist_var"]
                )
                task_item["vars"] = vars
                if include_type == "include":
                    b = task.get("task", {}).pop("become", None)
                    if b is not None:
                        task_item["vars"].setdefault("apply", {})["become"] = b
                    bu = task.get("task", {}).pop("become_user", None)
                    if bu is not None:
                        task_item["vars"].setdefault("apply", {})["become_user"] = bu

            elif task_type == "ansible-role":
                if command is None:
                    raise NsblException(
                        "No 'command' key specified in task: {}".format(task)
                    )
                include_type = task[self.task_marker]["include-type"]
                task_key = "{}_role".format(include_type)
                task_item[task_key] = {"name": command}
                task_item["vars"] = vars

                if include_type == "include":
                    b = task.get("task", {}).pop("become", None)
                    if b is not None:
                        task_item["vars"].setdefault("apply", {})["become"] = b
                    bu = task.get("task", {}).pop("become_user", None)
                    if bu is not None:
                        task_item["vars"].setdefault("apply", {})["become_user"] = bu

                for additional in [
                    "allow_duplicates",
                    "defaults_from",
                    "tasks_from",
                    "vars_from",
                ]:
                    value = task[self.task_marker].get(additional, None)
                    if value is not None:
                        task_item[task_key][additional] = value

            # add the remaining key/value pairs
            unknown_keys = []
            for key, value in task[self.task_marker].items():
                if key in ANSIBLE_TASK_KEYWORDS:
                    task_item[key] = value
                else:
                    unknown_keys.append(key)

            if task_has_password and not show_tasks_with_password_in_log:
                if "no_log" not in task_item.keys():
                    task_item["no_log"] = True

            # adding 'skip' conditions
            skip_internal_list = t[self.task_marker].get("__skip_internal__", [])
            if skip_internal_list:
                when_list = []
                when = task_item.get("when", None)
                if when:
                    when_list.append("( {} )".format(when))
                for skip_item in skip_internal_list:
                    when_list.append("( not {} )".format(skip_item))

                when_string = " and ".join(when_list)
                task_item["when"] = when_string

            result.append(task_item)

            reg_val = task_item.pop("register", None)
            if not reg_val:
                continue

            if not isinstance(reg_val, Mapping):
                temptask = copy.copy(task_item)
                temptask["register"] = reg_val
                raise FrklException(
                    msg="Can't parse task: {}".format(dict(temptask)),
                    reason="Invalid type for 'register' value: {}".format(
                        type(reg_val)
                    ),
                    solution="Use a mapping.",
                )

            register_var_name = reg_val["id"]
            task_item["register"] = register_var_name
            register_task = {
                "name": "[freckles register variable: {}]".format(register_var_name),
                "debug": {"msg": "{{{{ {} }}}}".format(register_var_name)},
            }
            result.append(register_task)

            # sys.exit()
            # if (
            #     "register" in task_item.keys()
            #     and not task_item["register"].startswith("__")
            #     and not task_item["register"].endswith("__")
            # ):
            #     register_name = task_item["register"]
            #     register_query = task[self.task_marker].get("register_query", None)
            #
            #     if register_query is not None:
            #         register_query_token = "::{}".format(register_query)
            #     else:
            #         register_query_token = ""
            #
            #     if "." in register_name:
            #         register_var_name, _ = register_name.split(".", 1)
            #         register_temp_name = "__temp_{}__".format(register_var_name)
            #         task_item["register"] = register_temp_name
            #
            #         set_fact_task = {
            #             "name": "[registering variable '{}' ]".format(register_name),
            #             "set_fact": {
            #                 "{}".format(
            #                     register_var_name
            #                 ): "{{{{ {} | register_var_filter('{}') }}}}".format(
            #                     register_temp_name, register_name
            #                 )
            #             },
            #         }
            #         register_task = {
            #             "name": "[freckles register variable: {}{}]".format(
            #                 register_var_name, register_query_token
            #             ),
            #             "debug": {"msg": "{{{{ {} }}}}".format(register_var_name)},
            #         }
            #         result.append(set_fact_task)
            #         result.append(register_task)
            #     else:
            #
            #         register_task = {
            #             "name": "[freckles register variable: {}{}]".format(
            #                 register_name, register_query_token
            #             ),
            #             "debug": {"msg": "{{{{ {} }}}}".format(register_name)},
            #         }
            #         result.append(register_task)

        return result

    def add_role_to_addition_files(self, command, path=None):

        if path is None:
            path = self.context.get_role_path(command)

        remote_role = path == command

        if not remote_role:
            if path not in self.additional_files:
                self.additional_files[path] = {
                    "type": ADD_TYPE_ROLE,
                    "target_name": command,
                }
        else:
            if command not in self.external_roles:
                self.external_roles.add(command)

    def add_tasklist_to_additional_files(self, tasklist_name, tasklist_var):

        if tasklist_var in self.additional_files.keys():
            return

        tasklist = self.context.get_tasklist(tasklist_name)

        self.additional_files[tasklist_var] = {
            "type": ADD_TYPE_TASK_LIST,
            "target_name": tasklist_var,
            "var_name": tasklist_var,
            "tasklist": tasklist,
        }



# -*- coding: utf-8 -*-

# python 3 compatibility
from __future__ import absolute_import, division, print_function, unicode_literals

import fnmatch
import logging
import os

from six import string_types

from frutils import dict_merge
from frutils.defaults import DEFAULT_EXCLUDE_DIRS

log = logging.getLogger("nsbl")

ROLE_CACHE = {}
ROLE_MARKER_FOLDERNAME = "meta"
ROLE_META_FILENAME = "main.yml"


def calculate_role_repos(role_repos):
    """Utility method to calculate which role repos to use.

    Role repos are folders containing ansible roles, and an (optional) task
    description file which is used to translate task-names in a task config
    file into roles or ansible tasks.

    Args:
      role_repos (list): a string or list of strings of local folders containing ansible roles

    Returns:
      list: a list of all local role repos to be used
    """

    if not role_repos:
        role_repos = []

    if isinstance(role_repos, string_types):
        role_repos = [role_repos]

    role_repos[:] = [os.path.realpath(os.path.expanduser(rr)) for rr in role_repos]

    role_repos[:] = [
        rr
        for rr in role_repos
        if os.path.exists(rr) and os.path.isdir(os.path.realpath(rr))
    ]
    log.debug("final role_repos: {}".format(role_repos))

    return role_repos


def find_roles_in_repos(role_repos):

    if isinstance(role_repos, string_types):
        role_repos = [role_repos]

    all_role_paths = set()
    result = {}
    for rr in role_repos:
        roles = find_roles_in_repo(rr)
        for v in roles.values():
            all_role_paths.add(v)
        # we want roles found first to take precedence
        result = dict_merge(roles, result, copy_dct=True)

    return result, all_role_paths


def find_roles_in_repo(role_repo):
    """Utility function to find all roles in a role_repo.

    Args:
      role_repo: the path to the role repo

    Returns:
    dict: a dictionary with the name of the role as key, and the path to the role as value
    """

    if role_repo in ROLE_CACHE.keys():
        return ROLE_CACHE[role_repo]

    result = {}
    try:

        for root, dirnames, filenames in os.walk(
            os.path.realpath(role_repo), topdown=True, followlinks=True
        ):

            dirnames[:] = [d for d in dirnames if d not in DEFAULT_EXCLUDE_DIRS]
            # check for meta folders
            for dirname in fnmatch.filter(dirnames, ROLE_MARKER_FOLDERNAME):

                meta_file = os.path.realpath(
                    os.path.join(root, dirname, ROLE_META_FILENAME)
                )
                if not os.path.exists(meta_file):
                    continue

                role_folder = root
                role_name = os.path.basename(role_folder)
                result[role_name] = role_folder

    except (UnicodeDecodeError):
        print(
            " X one or more filenames under '{}' can't be decoded, ignoring. This can cause problems later. ".format(
                root
            )
        )

    ROLE_CACHE[role_repo] = result

    return result



# -*- coding: utf-8 -*-

from __future__ import absolute_import, division, print_function, unicode_literals

import copy
import io
import sys

from jinja2 import Environment, FileSystemLoader
from ruamel.yaml.comments import CommentedMap
from six import string_types

from frkl import Frkl
from frkl.callbacks import FrklCallback
from frkl.defaults import (
    CHILD_MARKER_NAME,
    DEFAULT_LEAF_NAME,
    DEFAULT_LEAFKEY_NAME,
    OTHER_KEYS_NAME,
    KEY_MOVE_MAP_NAME,
)
from frkl.processors import (
    ConfigProcessor,
    UrlAbbrevProcessor,
    EnsureUrlProcessor,
    EnsurePythonObjectProcessor,
    FrklProcessor,
)
from frutils import StringYAML
from .defaults import *  # noqa: F403
from .exceptions import NsblException

NSBL_INVENTORY_BOOTSTRAP_FORMAT = {
    CHILD_MARKER_NAME: ENVS_KEY,
    DEFAULT_LEAF_NAME: ENV_META_KEY,
    DEFAULT_LEAFKEY_NAME: ENV_NAME_KEY,
    OTHER_KEYS_NAME: [VARS_KEY, TASKS_KEY],
    KEY_MOVE_MAP_NAME: VARS_KEY,
}

yaml = StringYAML()
yaml.default_flow_style = False
yaml.encoding = "utf-8"
yaml.allow_unicode = True


def parse_host_string(host_string):

    if not host_string:
        return {}

    username = None
    protocol = None
    host = None
    port = None

    if "://" in host_string:
        protocol, host = host_string.split("://")
    else:
        host = host_string

    if "@" in host:
        username, host = host.split("@")

    if ":" in host:
        host, port = host.split(":")

    result = {}
    if protocol:
        result["protocol"] = protocol

    if username:
        result["user"] = username

    if host:
        result["host"] = host

    if port:
        result["port"] = port

    return result


class NsblInventory(FrklCallback):
    def create(config, default_env_type=DEFAULT_ENV_TYPE, pre_chain=None):
        """Convenience method to create a NsblInventory object out of the configs and a few optional parameters.

        Args:
          config (list): list of config items
          default_env_type (str): the type a environment is if it is not explicitely specified, either ENV_TYPE_HOST or ENV_TYPE_GROUP
          pre_chain (list): the chain of ConfigProcessors to plug in front of the one that is used internally, needs to return a python list
        Returns:
          NsblInventory: the inventory object, already 'processed'
        """

        if pre_chain is None:
            pre_chain = [
                UrlAbbrevProcessor(),
                EnsureUrlProcessor(),
                EnsurePythonObjectProcessor(safe_load=False),
            ]

        chain = pre_chain + [FrklProcessor(**NSBL_INVENTORY_BOOTSTRAP_FORMAT)]
        inv_frkl = Frkl(config, chain)

        inventory = NsblInventory(default_env_type=default_env_type)

        inv_frkl.process(inventory)

        return inventory

    create = staticmethod(create)

    def __init__(self, **init_params):
        """Class to be used to create a dynamic ansible inventory from (elastic) yaml config files.

        The only init_params key that is supported is:

        default_env_type (str): the type a environment is if it is not explicitely specified, either ENV_TYPE_HOST or ENV_TYPE_GROUP
        Args:
          init_params (dict): the dict to initialize this object
        """
        super(NsblInventory, self).__init__(**init_params)
        self.groups = CommentedMap()
        self.hosts = CommentedMap()
        self.tasks = []
        self.current_env_id = 0

        self.localhost_only = True

        self.default_env_type = init_params.get("default_env_type", DEFAULT_ENV_TYPE)

    def result(self):
        return self.list()

    def extract_vars(self, inventory_dir):
        """Writes a folder structure with 'group_vars' and 'host_vars' folders into the target directory.

        Args:
          inventory_dir (str): the directory the inventory should be written to
        """
        for group, group_vars in self.groups.items():
            vars = group_vars.get(VARS_KEY, {})
            if not vars:
                continue
            group_dir = os.path.join(inventory_dir, "group_vars", group)
            var_file = os.path.join(group_dir, "{}.yml".format(group))
            # content = yaml.safe_dump(
            #    vars, default_flow_style=False, encoding="utf-8", allow_unicode=True
            # ).decode("utf-8")
            content = yaml.dump(vars).decode("utf-8")

            os.makedirs(group_dir)
            with io.open(var_file, "w", encoding="utf-8") as text_file:
                text_file.write(content)

        for host, host_vars in self.hosts.items():
            vars = host_vars.get(VARS_KEY, {})
            if not vars:
                continue
            host_dir = os.path.join(inventory_dir, "host_vars", host)
            var_file = os.path.join(host_dir, "{}.yml".format(host))
            # content = yaml.safe_dump(
            #    vars, default_flow_style=False, encoding="utf-8", allow_unicode=True
            # ).decode("utf-8")
            content = yaml.dump(vars)
            if hasattr(content, "decode"):
                content = content.decode("utf-8")

            os.makedirs(host_dir)
            with io.open(var_file, "w", encoding="utf-8") as text_file:
                text_file.write(content)

    def get_inventory_config_string(self):
        """Returns a string that can be used to write an ansible hosts file, including hosts, groups and child-groups."""

        jinja_env = Environment(loader=FileSystemLoader(TEMPLATES_FOLDER))
        # jinja_env = Environment(loader=PackageLoader("nsbl", "templates"))
        template = jinja_env.get_template("hosts")
        output_text = template.render(groups=self.groups, hosts=self.hosts)

        return output_text

    def write_inventory_file_or_script(
        self, inventory_dir, extract_vars=False, relative_paths=True
    ):
        """Writes an ansible hosts file or dynamic inventory script into the provided directory.

        Writing a dynamic inventory script is not implemented yet.

        Args:
          inventory_dir (str): the target directory
          extract_vars (bool): whether to extract all vars (True, default) or write a dynamic inventory script
          relative_paths (bool): only important for when writing dynamic inventory scripts, makes the paths in the script relative to the ansible environment root so its easily copy-able
        """
        if extract_vars:
            inventory_string = self.get_inventory_config_string()
            inventory_name = "hosts"
            inventory_file = os.path.join(inventory_dir, inventory_name)
            par_dir = os.path.dirname(inventory_file)
            if not os.path.exists(par_dir):
                os.makedirs(par_dir)

            with io.open(inventory_file, "w", encoding="utf-8") as text_file:
                text_file.write(inventory_string)

        else:
            raise Exception("Dynamic inventory script creation not implemented yet.")
            # else:
            # # write dynamic inventory script
            # jinja_env = Environment(loader=PackageLoader('nsbl', 'templates'))
            # if relative_paths:
            #     template = jinja_env.get_template('inventory_relative')
            # else:
            #     template = jinja_env.get_template('inventory_absolute')

            # if relative_paths:
            #     for path in self.configs:
            #         rel_path = os.path.relpath(path, inventory_dir)
            #         rel_configs.append(rel_path)

            #     script_configs = " --config ".join(rel_configs)
            # else:
            #     abs_configs = []
            #     for path in self.configs:
            #         abs_path = os.path.abspath(path)
            #         abs_configs.append(abs_path)
            #     script_configs = " --config".join(abs_configs)

            # output_text = template.render(role_repo_paths=roles_repos_string, nsbl_script_configs=script_configs)

            # inventory_string = self.get_inventory_config_string()
            # inventory_target_name = "inventory"

            # inventory_file = os.path.join(inventory_dir, inventory_target_name)

            # with open(inventory_file, "w") as text_file:
            #     text_file.write(output_text)

            # st = os.stat(inventory_file)
            # os.chmod(inventory_file, 0o775)
            # pass

    def add_group(self, group_name, group_vars):
        """Add a group to the dynamic inventory.

        Args:
          group_name (str): the name of the group
          group_vars (dict): the variables for this group
        """

        if group_name in self.groups.keys():
            raise NsblException("Group '{}' defined twice".format(group_name))

        self.groups[group_name] = {}
        self.groups[group_name]["vars"] = group_vars
        self.groups[group_name]["hosts"] = []
        self.groups[group_name]["children"] = []

    def add_host(self, host_name, host_vars):
        """Add a host to the dynamic inventory.

        Args:
          host_name (str): the name of the host
          host_vars (dict): the variables for this host
        """

        if host_name not in self.hosts.keys():
            self.hosts[host_name] = {VARS_KEY: {}}

        if not host_vars:
            return

        intersection = set(self.hosts[host_name].get(VARS_KEY, {}).keys()).intersection(
            host_vars.keys()
        )

        if intersection:
            raise NsblException(
                "Adding host more than once with intersecting keys, this is not possible because it's not clear which vars should take precedence. Intersection: {}".format(
                    intersection
                )
            )

        self.hosts[host_name][VARS_KEY].update(host_vars)

    def add_group_to_group(self, child, group):
        """Adds a group as a subgroup of another group.

        Args:
          child (str): the name of the sub-group
          group (str): the name of the parent group
        """

        if group not in self.groups[group]["children"]:
            self.groups[group]["children"].append(child)

    def add_host_to_group(self, host, group):
        """Adds a host to a group.

        Args:
          host (str): the name of the host
          group (str): the name of the parent group
        """

        if host not in self.groups[group]["hosts"]:
            self.groups[group]["hosts"].append(host)

        self.add_host(host, None)

    def callback(self, env):
        """Adds a new environment, and sorts it into the appropriate internal variable."""

        if ENV_META_KEY not in env.keys():
            raise NsblException(
                "Environment does not have metadata (missing '{}') key: {})".format(
                    ENV_META_KEY, env
                )
            )
        env_type = env[ENV_META_KEY].get(ENV_TYPE_KEY, False)
        if not env_type:
            if ENV_HOSTS_KEY in env[ENV_META_KEY].keys():
                env_type = ENV_TYPE_GROUP
            elif ENV_GROUPS_KEY in env[ENV_META_KEY].keys():
                env_type = ENV_TYPE_HOST
            else:
                env_type = self.default_env_type

        env_name = env[ENV_META_KEY].get(ENV_NAME_KEY, False)
        if not env_name:
            raise NsblException(
                "Environment metadata needs to contain a name (either host- or group-name): {}".format(
                    env[ENV_META_KEY]
                )
            )

        if env_type == ENV_TYPE_HOST:

            self.add_host(env_name, env.get(VARS_KEY, {}))

            if ENV_HOSTS_KEY in env.get(ENV_META_KEY, {}).keys():
                raise NsblException(
                    "An environment of type {} can't contain the {} key".format(
                        ENV_TYPE_HOST, ENV_HOSTS_KEY
                    )
                )

            for group in env[ENV_META_KEY].get(ENV_GROUPS_KEY, []):
                self.add_host_to_group(env_name, group)

        elif env_type == ENV_TYPE_GROUP:

            self.add_group(env_name, env.get(VARS_KEY, {}))

            for host in env[ENV_META_KEY].get(ENV_HOSTS_KEY, []):
                self.add_host_to_group(host, env_name)

            for group in env[ENV_META_KEY].get(ENV_GROUPS_KEY, []):
                self.add_group_to_group(group, env_name)

        else:
            raise NsblException(
                "Environment type needs to be either 'host' or 'group': {}".format(
                    env_type
                )
            )

        if TASKS_KEY in env.keys():

            current_meta = copy.deepcopy(env[ENV_META_KEY])
            current_meta[ENV_ID_KEY] = self.current_env_id
            env_name = env[ENV_META_KEY].get(ENV_NAME_KEY, False)
            if not env_name:
                raise NsblException(
                    "Environment metadata needs to contain a name (either host- or group-name): {}".format(
                        env[ENV_META_KEY]
                    )
                )
            current_meta[ENV_NAME_KEY] = env_name
            self.tasks.append(
                {
                    TASKS_META_KEY: current_meta,
                    TASKS_KEY: env[TASKS_KEY],
                    VARS_KEY: env.get(VARS_KEY, {}),
                }
            )
            self.current_env_id += 1

    def finished(self):
        if (
            "localhost" in self.hosts.keys()
            and "ansible_connection"
            not in self.hosts["localhost"].get(VARS_KEY, {}).keys()
        ):
            self.hosts["localhost"][VARS_KEY]["ansible_connection"] = "local"

        if (
            "localhost" in self.hosts.keys()
            and "ansible_python_interpreter"
            not in self.hosts["localhost"].get(VARS_KEY, {}).keys()
        ):

            if not os.path.exists("/usr/bin/python"):
                self.hosts["localhost"][VARS_KEY][
                    "ansible_python_interpreter"
                ] = sys.executable

        for host in self.hosts.keys():
            if host != "localhost" and host != "127.0.0.1":
                self.localhost_only = False
                break

            connection = (
                self.hosts[host].get(VARS_KEY, {}).get("ansible_connection", None)
            )
            if connection != "local":
                self.localhost_only = False
                break

    def list(self):
        """Lists all groups in the format that is required for ansible dynamic inventories.

        More info: https://docs.ansible.com/ansible/intro_dynamic_inventory.html, http://docs.ansible.com/ansible/dev_guide/developing_inventory.html

        Returns:
        dict: a dict containing all information about all hosts/groups
        """

        result = copy.deepcopy(self.groups)
        result["_meta"] = {"hostvars": {}}
        for host, vars in self.hosts.items():
            result["_meta"]["hostvars"][host] = vars.get(VARS_KEY, {})

        # return json.dumps(result, sort_keys=4, indent=4)
        return result

    def host(self, host):
        """Returns the inventory information for the specified host, in the format required for ansible dynamic inventories.

        Args:
          host (str): the name of the host
        Returns:
        dict: all inventory information for this host
        """

        host_vars = self.hosts.get(host, {}).get(VARS_KEY, {})
        return host_vars

    def get_vars(self, env_name):
        """Returns all variables for the environment with the specified name.

        First tries whether the name matches a group, then tries hosts.

        Args:
          env_name (str): the name of the group or host
        Returns:
          dict: the variables for the environment
        """

        if env_name in self.groups.keys():
            return self.groups[env_name].get(VARS_KEY, {})
        elif env_name in self.hosts.keys():
            return self.hosts[env_name].get(VARS_KEY, {})
        else:
            raise NsblException(
                "Neither group or host with name '{}' exists".format(env_name)
            )


class WrapTasksIntoLocalhostEnvProcessor(ConfigProcessor):
    """Wraps a list of tasks into a localhost environment.

    Convenience processor to not have to do this manually, keeps configuration files minimal and sweet.
    """

    def __init__(self, init_params=None):
        super(WrapTasksIntoLocalhostEnvProcessor, self).__init__(init_params)

        self.task_configs = []

    def validate_init(self):

        self.task_vars = self.init_params.get(VARS_KEY, {})

        return True

    def handles_last_call(self):

        return True

    def process_current_config(self):

        config = self.current_input_config

        if not self.last_call:
            self.task_configs.append(config)
        else:
            result = {
                "localhost": {
                    TASKS_KEY: self.task_configs,
                    TASKS_META_KEY: {ENV_TYPE_KEY: ENV_TYPE_HOST},
                    VARS_KEY: self.task_vars,
                }
            }

            return result


class WrapTasksIntoHostsProcessor(ConfigProcessor):
    """Wraps a list of tasks into a localhost environment.

    Convenience processor to not have to do this manually, keeps configuration files minimal and sweet.
    """

    def __init__(self, init_params=None):
        super(WrapTasksIntoHostsProcessor, self).__init__(init_params)

        self.task_configs = []

    def validate_init(self):

        self.task_vars = self.init_params.get(VARS_KEY, {})
        self.hosts = self.init_params.get(ENV_HOSTS_KEY, [])

        return True

    def handles_last_call(self):

        return True

    def process_current_config(self):

        config = self.current_input_config

        if not self.last_call:
            self.task_configs.append(config)
        else:

            result = []

            for host in self.hosts:

                if isinstance(host, string_types):
                    details = parse_host_string(host)
                elif isinstance(host, dict):
                    details = host
                else:
                    raise Exception(
                        "Can't parse host, unknown type (can only be string or dict): {}".format(
                            host
                        )
                    )

                temp_vars = copy.deepcopy(self.task_vars)

                if "host" in details.keys():
                    temp_vars["host"] = details["host"]
                else:
                    temp_vars["host"] = "localhost"

                if "user" in details.keys():
                    temp_vars["ansible_user"] = details["user"]

                if "protocol" in details.keys():
                    temp_vars["ansible_connection"] = details["protocol"]
                else:
                    if (
                        temp_vars["host"] == "localhost"
                        or temp_vars["host"] == "127.0.0.1"
                    ):
                        temp_vars["ansible_connection"] = "local"
                    else:
                        temp_vars["ansible_connection"] = "ssh"

                if details.get("port", 0) > 0:
                    temp_vars["ansible_port"] = details["port"]

                temp = {
                    temp_vars["host"]: {
                        TASKS_KEY: self.task_configs,
                        TASKS_META_KEY: {ENV_TYPE_KEY: ENV_TYPE_HOST},
                        VARS_KEY: temp_vars,
                    }
                }
                result.append(temp)

            return result
